# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

import asyncio
from asyncio.events import get_event_loop
from contextlib import AbstractAsyncContextManager
import xml.etree.ElementTree as ET

from aioraven.device import RAVEnBaseDevice
from aioraven.protocols import RAVEnReaderProtocol


async def open_connection(host=None, port=None, *, loop=None, **kwargs):
    """
    Establish a network connection to a RAVEn device.

    Additional optional keyword arguments are passed to
    `AbstractEventLoop.create_connection()`.

    :param host: The hostname or IP address to connect to.
    :param port: The TCP port number to connect to.
    :param loop: The event loop instance to use.
    """
    if loop is None:
        loop = get_event_loop()
    reader = RAVEnReader(loop=loop)
    protocol = RAVEnReaderProtocol(reader, loop=loop)
    transport, _ = await loop.create_connection(
        lambda: protocol, host=host, port=port, **kwargs)
    writer = RAVEnWriter(transport, protocol)
    return reader, writer


class RAVEnReader:
    """Read stanzas from a RAVEn device."""

    def __init__(self, loop=None):
        """
        Construct a RAVEnReader.

        :param loop: The event loop instance to use.
        """
        if loop is None:
            self._loop = get_event_loop()
        else:
            self._loop = loop
        self._eof = False
        self._waiters = {None: []}
        self._exception = None

    def __repr__(self):
        info = [self.__class__.__name__]
        if self._eof:
            info.append('eof')
        if self._exception:
            info.append('e=%r' % self._exception)
        num_waiters = sum(len(w) for w in self._waiters.values())
        if num_waiters:
            info.append('w=%d' % num_waiters)
        return '<%s>' % ' '.join(info)

    def exception(self):
        return self._exception

    def set_exception(self, exc):
        # Only ParseError is recoverable
        if not isinstance(exc, ET.ParseError):
            self._exception = exc
        for waiters in self._waiters.values():
            while waiters:
                waiter = waiters.pop(0)
                if not waiter.cancelled():
                    waiter.set_exception(exc)

    def feed_eof(self):
        self._eof = True
        for waiters in self._waiters.values():
            while waiters:
                waiter = waiters.pop(0)
                if not waiter.cancelled():
                    waiter.set_result(None)

    def feed_element(self, data):
        self._waiters.setdefault(data.tag, [])
        waiters = self._waiters.get(data.tag, [])
        res = {}
        for e in data:
            if e.tag in res:
                if not isinstance(res[e.tag], list):
                    res[e.tag] = [res[e.tag]]
                res[e.tag].append(e.text)
            else:
                res[e.tag] = e.text
        while waiters:
            waiter = waiters.pop(0)
            if not waiter.cancelled():
                waiter.set_result(res)
                break
        else:
            waiters = self._waiters[None]
            if waiters:
                waiters.pop(0).set_result(res)

    async def read_tag(self, tag=None):
        if self._eof:
            return None
        if self._exception is not None:
            raise self._exception
        self._waiters.setdefault(tag, [])
        waiter = self._loop.create_future()
        self._waiters[tag].append(waiter)
        return await waiter


class RAVEnWriter:
    """Write commands to a RAVEn device."""

    def __init__(self, transport, protocol):
        """
        Construct a RAVEnWriter.

        :param transport: The transport instance to wrap.
        :param protocol: The reader protocol for the connection.
        """
        self._transport = transport
        self._protocol = protocol

    def __repr__(self):
        info = [self.__class__.__name__]
        info.append('transport=%r' % self._transport)
        return '<%s>' % ' '.join(info)

    def write_cmd(self, cmd_name, args=None):
        element_cmd = ET.Element('Command')
        element_name = ET.SubElement(element_cmd, 'Name')
        element_name.text = cmd_name
        for k, v in (args or {}).items():
            element_arg = ET.SubElement(element_cmd, k)
            element_arg.text = v
        tree = ET.ElementTree(element_cmd)
        tree.write(self._transport, encoding='ASCII', xml_declaration=False)

    def close(self):
        return self._transport.close()

    async def wait_closed(self):
        await self._protocol._get_close_waiter(self)


class RAVEnStreamDevice(RAVEnBaseDevice, AbstractAsyncContextManager):
    """Read and write coordination for stream-based RAVEn devices."""

    _reader = None
    _writer = None

    async def open(self):
        raise NotImplementedError()

    async def close(self):
        if self._writer:
            self._writer.close()
            try:
                await self._writer.wait_closed()
            except IOError:
                pass
            self._reader = None
            self._writer = None

    async def _query(self, cmd_name, res_name=None, args=None):
        if not self._reader or not self._writer:
            raise RuntimeError('Device is not open')
        waiter = None
        if res_name:
            waiter = asyncio.create_task(self._reader.read_tag(res_name))
            await asyncio.sleep(0)
        self._writer.write_cmd(cmd_name, args)
        return await waiter if waiter else None

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()


class RAVEnNetworkDevice(RAVEnStreamDevice):
    """A network-connected RAVEn device."""

    def __init__(self, host=None, port=None, *, loop=None, **kwargs):
        """
        Construct a RAVEnNetworkDevice.

        Additional optional keyword arguments are passed to
        `AbstractEventLoop.create_connection()`.

        :param host: The hostname or IP address to connect to.
        :param port: The TCP port number to connect to.
        :param loop: The event loop instance to use.
        """
        self._host = host
        self._port = port
        self._loop = loop
        self._kwargs = kwargs

    def __repr__(self):
        info = [self.__class__.__name__]
        info.append('host=%s' % self._host)
        info.append('port=%s' % self._port)
        return '<%s>' % ' '.join(info)

    async def open(self):
        """Open the connection to the RAVEn device."""
        if self._reader or self._writer:
            return
        self._reader, self._writer = await open_connection(
            host=self._host, port=self._port, loop=self._loop, **self._kwargs)



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

import asyncio
import os
import sys
import xml.etree.ElementTree as ET

from aioraven.data import MeterList
from aioraven.serial import RAVEnSerialDevice
import pytest
from serial import SerialException

from .mock_device import mock_device


if os.name == "nt":
    pytest.skip(
        "Socket serial protocol is currently broken with pyserial_asyncio",
        allow_module_level=True)


@pytest.mark.asyncio
async def test_serial_data():
    """Verify simple device query behavior."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnSerialDevice(f'socket://{host}:{port}') as dut:
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[])


@pytest.mark.asyncio
@pytest.mark.skipif(
    sys.platform == 'darwin',
    reason="Socket serial protocol doesn't support EOF")
async def test_serial_disconnect():
    """Verify behavior when a device is unexpectedly disconnected."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        dut = RAVEnSerialDevice(f'socket://{host}:{port}')
        await dut.open()
        assert await dut.get_meter_list()
    async with dut:
        with pytest.raises(SerialException):
            await dut.get_meter_list()
        with pytest.raises(SerialException):
            await dut.get_meter_list()


@pytest.mark.asyncio
@pytest.mark.skipif(
    sys.platform == 'darwin',
    reason="Socket serial protocol doesn't support EOF")
async def test_serial_incomplete():
    """Verify behavior when a partial fragment is received."""
    responses = {
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>'
            b'<DeviceInfo>'
    }

    async with mock_device(responses) as (host, port):
        dut = RAVEnSerialDevice(f'socket://{host}:{port}')
        await dut.open()
        assert await dut.get_meter_list()

        task = asyncio.create_task(dut.get_meter_list())
        await asyncio.wait((task,), timeout=0.05)
    async with dut:
        # Pyserial doesn't get EOF from sockets, so we won't see the
        # ET.ParseError we're expecting here. Rather, we'll see a
        # SerialException stating that the socket was disconnected
        # (which isn't entirely correct).
        with pytest.raises(SerialException):
            assert not await task


@pytest.mark.asyncio
async def test_serial_not_open():
    """Verify behavior when reading from an unopened device."""
    async with mock_device() as (host, port):
        dut = RAVEnSerialDevice(f'socket://{host}:{port}')
        with pytest.raises(RuntimeError):
            await dut.get_device_info()


@pytest.mark.asyncio
async def test_serial_parse_error():
    """Verify behavior when invalid syntax is received."""
    responses = {
        b'<Command><Name>get_device_info</Name></Command>':
            b'</DeviceInfo>',
        b'<Command><Name>get_meter_list</Name></Command>':
            b'<MeterList>'
            b'    <DeviceMacId>0x0123456789abcdef</DeviceMacId>'
            b'</MeterList>',
    }

    async with mock_device(responses) as (host, port):
        async with RAVEnSerialDevice(f'socket://{host}:{port}') as dut:
            with pytest.raises(ET.ParseError):
                await dut.get_device_info()
            actual = await dut.get_meter_list()

    assert actual == MeterList(
        device_mac_id=bytes.fromhex('0123456789ABCDEF'),
        meter_mac_ids=[])


@pytest.mark.asyncio
async def test_serial_repr():
    """Verify representation of a serial device."""
    async with mock_device({}) as (host, port):
        async with RAVEnSerialDevice(f'socket://{host}:{port}') as dut:
            assert 'RAVEnSerialDevice' in str(dut)



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

from asyncio.events import get_event_loop
from asyncio.protocols import Protocol
import warnings
import xml.etree.ElementTree as ET


class DeviceWarning(Warning):
    """The device has generated a warning message."""

    def __init__(self, message):
        """
        Construct a DeviceWarning.

        :param message: The reason for which the device is warning us.
        """
        super().__init__('Warning from RAVEn device: ' + message)


class UnknownCommandWarning(DeviceWarning):
    """A recently executed command is not supported by the device."""

    MESSAGE = 'Unknown command'

    def __init__(self):
        """Construct a UnknownCommandWarning."""
        super().__init__(self.MESSAGE)


class RAVEnReaderProtocol(Protocol):
    """Deserialize data fragments from a RAVEn device."""

    def __init__(self, reader, loop=None):
        """
        Construct a RAVEnRaderProtocol.

        :param reader: The `RAVEnReader` instance to which deserialized
          fragments are passed.
        :param loop: The event loop instance to use.
        """
        if loop is None:
            self._loop = get_event_loop()
        else:
            self._loop = loop
        self._reader = reader
        self._parser = None
        self._closed = self._loop.create_future()

    def _reset(self):
        self._parser = ET.XMLPullParser(events=('end',))
        self._parser.feed(b'<?xml version="1.0" encoding="ASCII"?><root>')

    def _get_close_waiter(self, stream):
        return self._closed

    def connection_made(self, transport):
        self._reset()

    def connection_lost(self, exc):
        if self._reader is not None:
            if exc is None:
                self._reader.feed_eof()
            else:
                self._reader.set_exception(exc)
        if not self._closed.done():
            if exc is None:
                self._closed.set_result(None)
            else:
                self._closed.set_exception(exc)
        self._reader = None
        self._parser = None

    def data_received(self, data):
        self._parser.feed(data)

        events = self._parser.read_events()
        while True:
            try:
                _, element = next(events)
            except StopIteration:
                return
            except ET.ParseError as e:
                self._reader.set_exception(e)
                self._reset()
            else:
                if element.tag == 'Warning':
                    try:
                        e = next(iter(element), None)
                        text = e.text if e is not None else 'Unknown warning'
                        if text == UnknownCommandWarning.MESSAGE:
                            warnings.warn(UnknownCommandWarning())
                        else:
                            warnings.warn(DeviceWarning(text))
                    except Warning as e:
                        self._reader.set_exception(e)
                else:
                    self._reader.feed_element(element)

    def eof_received(self):
        self._parser.feed(b'</root>')

        try:
            self._parser.close()
        except ET.ParseError as e:
            self._reader.set_exception(e)
        finally:
            self._reader.feed_eof()

    def __del__(self):
        try:
            closed = self._closed
        except AttributeError:
            pass
        else:
            if closed.done() and not closed.cancelled():
                closed.exception()



# Copyright 2022 Scott K Logan
# Licensed under the Apache License, Version 2.0

from dataclasses import dataclass
from datetime import date
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from enum import Enum
from enum import IntEnum
from typing import List
from typing import Tuple

from iso4217 import Currency


def convert_bool(raw):
    if raw is None:
        return None
    if raw == 'Y':
        return True
    if raw == 'N':
        return False
    raise ValueError(f"Invalid boolean format: '{raw}'")


def convert_currency(raw):
    if raw is None:
        return None
    currency = next(filter(lambda c: c.number == int(raw, 0), Currency), None)
    if currency is None:
        raise ValueError(f"Invalid currency number: '{raw}'")
    return currency


def convert_date(raw):
    if raw is None:
        return None
    if len(raw) < 8:
        raise ValueError(f"Invalid date format: '{raw}'")
    return date(int(raw[:4]), int(raw[4:6]), int(raw[6:8]))


def convert_date_code(raw):
    if raw is None:
        return None
    if len(raw) < 9:
        raise ValueError(f"Invalid date code format: '{raw}'")
    return (
        convert_date(raw[:8]),
        raw[8:],
    )


def convert_datetime(raw, utc=False):
    if raw is None or raw == 0xffffff:
        return None
    value = datetime(
        2000, 1, 1, 0, 0,
        tzinfo=timezone.utc if utc else None)
    return value + timedelta(seconds=int(raw, 0))


def convert_float(raw_value, raw_div):
    if raw_value is None:
        return None
    value = float(int(raw_value, 0))
    if raw_div is None:
        return value
    return value / 10 ** int(raw_div, 0)


def convert_hex_to_bytes(raw):
    if raw is None:
        return None
    if raw.startswith('0x'):
        raw = raw[2:]
    return bytes.fromhex(raw)


def convert_int(raw):
    if raw is None:
        return raw
    return int(raw, 0)


def convert_int_formatted(
    raw, mult=None, div=None, digits_right=None, digits_left=None,
    suppress_leading_zero=None
):
    if raw is None:
        return None
    value = int(raw, 0)
    if mult is not None:
        mult = int(mult, 0)
        value *= mult
    if div is not None:
        div = int(div, 0)
        value /= div
    precision = ""
    if digits_right is not None:
        precision = int(digits_right, 0)
    width = ""
    if digits_left is not None:
        if precision == "":
            precision = len(str(value).split(".")[-1])
        width = int(digits_left, 0) + 1 + precision
    value = "{:0>{width}.{precision}f}".format(value, **locals())
    if suppress_leading_zero is not None:
        if suppress_leading_zero == "Y":
            value = value.lstrip("0")
            if value[0] == ".":
                value = "0" + value
    return value


def convert_price(raw_value, raw_div, exponent):
    value = str(convert_float(raw_value, raw_div))
    if exponent is not None:
        major, minor = value.split('.', 1)
        if exponent == 0 and minor == '0':
            return major
        missing = exponent - len(minor)
        if missing > 0:
            value += '0' * missing
    return value


def convert_timedelta(raw):
    if raw is None:
        return None
    value = int(raw, 0)
    return timedelta(seconds=value)


class ConnectionState(str, Enum):
    """Indicates the current state of the device."""

    __str__ = str.__str__

    INITIALIZING = "Initializing"
    NETWORK_DISCOVERY = "Network Discovery"
    JOINING = "Joining"
    JOIN_FAIL = "Join: Fail"
    JOIN_SUCCESS = "Join: Success"
    AUTHENTICATING = "Authenticating"
    AUTHENTICATING_SUCCESS = "Authenticating: Success"
    AUTHENTICATING_FAIL = "Authenticating: Fail"
    CONNECTED = "Connected"
    DISCONNECTED = "Disconnected"
    REJOINING = "Rejoining"


class DataStatus(IntEnum):
    """Indicates the status of returned data."""

    SUCCESS = 0
    UNDEFINED = 1
    NOT_SUPPORTED = 2
    INVALID_END_TIME = 3
    TOO_MANY = 4
    NONE_AVAILABLE = 5


class IntervalChannel(str, Enum):
    """Profile interval data channels."""

    DELIVERED = "Delivered"
    RECEIVED = "Received"


class IntervalPeriod(IntEnum):
    """Sampling intervals."""

    DAILY = 0
    SIXTY_MIN = 1
    THIRTY_MIN = 2
    FIFTEEN_MIN = 3
    TEN_MIN = 4
    SEVEN_AND_A_HALF_MIN = 5
    FIVE_MIN = 6
    TWO_AND_A_HALF_MIN = 7


class MessageQueue(str, Enum):
    """Types of queues messages can occupy."""

    __str__ = str.__str__

    ACTIVE = "Active"
    CANCEL_PENDING = "Cancel Pending"


class MeterType(str, Enum):
    """Types of meters to connect to."""

    __str__ = str.__str__

    ELECTRIC = "electric"
    GAS = "gas"
    WATER = "water"
    OTHER = "other"


class ScheduledEvent(str, Enum):
    """Types of events which can be scheduled."""

    __str__ = str.__str__

    TIME = "time"
    PRICE = "price"
    DEMAND = "demand"
    SUMMATION = "summation"
    MESSAGE = "message"


@dataclass
class ConnectionStatus:
    """Diagnostic information about the meter connection."""

    device_mac_id: bytes
    coord_mac_id: bytes
    status: ConnectionState
    description: str
    status_code: bytes
    ext_pan_id: bytes
    channel: int
    short_addr: bytes
    link_strength: int


@dataclass
class CurrentPeriodUsage:
    """Total consumption for current accumulation period."""

    device_mac_id: bytes
    meter_mac_id: bytes
    time_stamp: datetime
    current_usage: str
    start_date: datetime


@dataclass
class CurrentSummationDelivered:
    """Total consumption at the meter to date."""

    device_mac_id: bytes
    meter_mac_id: bytes
    time_stamp: datetime
    summation_delivered: str
    summation_received: str


@dataclass
class DeviceInfo:
    """Information about the device."""

    device_mac_id: bytes
    install_code: bytes
    link_key: bytes
    fw_version: str
    hw_version: str
    image_type: str
    manufacturer: str
    model_id: str
    date_code: Tuple[date, str]


@dataclass
class InstantaneousDemand:
    """Current consumption rate at the meter."""

    device_mac_id: bytes
    meter_mac_id: bytes
    time_stamp: datetime
    demand: str


@dataclass
class LastPeriodUsage:
    """Total consumption for the previous accumulation period."""

    device_mac_id: bytes
    meter_mac_id: bytes
    last_usage: str
    start_date: datetime
    end_date: datetime


@dataclass
class MessageCluster:
    """Text messages from the meter."""

    device_mac_id: bytes
    meter_mac_id: bytes
    time_stamp: datetime
    message_id: bytes
    text: str
    confirmation_required: bool
    confirmed: bool
    queue: MessageQueue


@dataclass
class MeterInfo:
    """Information about a meter on the network."""

    device_mac_id: bytes
    meter_mac_id: bytes
    meter_type: MeterType
    nick_name: str
    account: str
    auth: str
    host: str
    enabled: bool


@dataclass
class MeterList:
    """List of meters the device is connected to."""

    device_mac_id: bytes
    meter_mac_ids: List[bytes]


@dataclass
class NetworkInfo(ConnectionStatus):
    """Information about the network the device is on."""

    device_mac_id: bytes
    coord_mac_id: bytes
    status: ConnectionState
    description: str
    status_code: bytes
    ext_pan_id: bytes
    channel: int
    short_addr: bytes
    link_strength: int


@dataclass
class PriceCluster:
    """The current price in effect on the meter."""

    device_mac_id: bytes
    meter_mac_id: bytes
    time_stamp: datetime
    price: str
    currency: Currency
    tier: int
    tier_label: str
    rate_label: str
    # TODO(cottsay): Add 'Duration'


@dataclass
class ProfileData:
    """A series of interval data as recorded by the meter."""

    device_mac_id: bytes
    meter_mac_id: bytes
    end_time: datetime
    status: DataStatus
    profile_interval_period: IntervalPeriod


@dataclass
class ScheduleInfo:
    """Information about periodic notifications sent by the device."""

    device_mac_id: bytes
    meter_mac_id: bytes
    event: ScheduledEvent
    frequency: timedelta
    enabled: bool


@dataclass
class TimeCluster:
    """The current time reported on the meter."""

    device_mac_id: bytes
    meter_mac_id: bytes
    utc_time: datetime
    local_time: datetime
